// format.C Version 7.40  Copyleft (c) 杉本典夫
// Time-stamp: <2022-11-29 16:14:13 tac>
//
// 書式クラスのメンバー関数
//
// ライブラリ:libdans.so.7.4
// コンパイル法:g++ -O -fPIC -c format.C

#include <dans.H>               // dans用ヘッダファイル(fun.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

namespace                       // 名前無し名前空間(局所関数の定義)
{

strsize_t getsubfmt(string& substr, const string& sstr, char delmc)
// 区切り文字までの書式指定を取り出す
// ()中または//中に区切り記号があっても無視する。
{
  if (sstr.empty()) return string::npos;
  substr.erase();
  bool str = false, bracket = false;
  char cpre = '\0';
  strsize_t i = 0;
  string tstr(sstr), buf;
  int nc = 0;
  if (delmc == ' ') while (i < tstr.size() && tstr[i] == delmc) ++i;
  while (i < tstr.size())
  {
    char c = tstr[i];
    if (nc > 1)                 // マルチバイトコードの途中
    { buf += c;  --nc; }
    else if (nc == 1)           // マルチバイトコードの最後
    {
      buf += c;
      substr += buf;
      buf.erase();  nc = 0;
    }
    else if (ismcode(c))        // マルチバイトコード1バイト目
    { buf = c;  nc = ncode(c) - 1; }
    else if (cpre == CHResc)    // \<文字>
    { substr += c;  cpre = '\0'; }
    else if (bracket)           // 括弧中
    {
      substr += c;
      if (c == ')') bracket = false; // 括弧終了
    }
    else if (str)               // 文字列中
    {
      substr += c;
      if (c == CHRstr) str = false; // 文字列終了
    }
    else if (c == CHResc)       // エスケープ文字「\」
    { substr += c;  cpre = c; }
    else if (c == '(')          // 括弧開始
    { substr += c;  bracket = true; }
    else if (c == CHRstr)       // 文字列開始
    { substr += c;  str = true; }
    else if (c == delmc)        // 区切り文字「,」
    { break; }
    else                        // その他の文字
    { substr += c; }
    ++i;
  }
  return i;                     // 区切り記号のオフセットを返す
}

int fmtsplit(lstr& dstrlst, const string& sstr, char delmc)
// 書式指定を区切り文字で分割して文字列リストに追加する
{
  if (sstr.empty()) return 0;   // 文字列が空の場合
  string tstr(sstr), substr;
  if (delmc == ' ') strtrm(tstr);
  int size0 = dstrlst.size();
  strsize_t idx = string::npos;
  while ((idx = getsubfmt(substr, tstr, delmc)) != string::npos)
  {
    dstrlst.push_back(substr);
    if (idx < tstr.size()) tstr.erase(0, idx + 1);
    else break;                 // 文字列終了
    if (delmc == ' ') strtrm(tstr);
  }
  if (idx == string::npos) dstrlst.push_back(""); // 最後に区切り文字がある場合
  return dstrlst.size() - size0; // 追加した要素数を返す
}

} // 名前無し名前空間終わり

const format& format::operator=(const format& format0)
// コピー代入演算子
{
  if (&format0 == this) return *this; // 自分自身への代入を防ぐ
  vnf = format0.vnf;  vstr = format0.vstr;
  return *this;
}

void format::clear()            // 書式を削除する
{
  vnf.first = 0;  vnf.second = -1;  vstr.clear();
}

int format::getfmt(const string& str) // 文字列から書式を取り出す
{
  if (str.empty()) return -1;   // 文字列不適
  vnf.first = 0;  vnf.second = -1;  vstr.erase();
  string tstr(str), substr;
  if (tstr[0] == CHRstr)        // 出力文字列指定
  {
    tstr.erase(0, 1);
    strsize_t idx = getsubstr(substr, tstr, CHRstr);
    if (idx == string::npos || idx >= tstr.size() || substr.empty()) return -2;
    sstrtostr(substr);          // 特殊表現文字を変換し""を削除する
    vstr = substr;
    vnf.first = strclen(vstr, CHARwide); // 表示桁数
  }
  else                          // 桁指定
  {
    char* ptr = 0;
    strsize_t idx = getsubstr(substr, tstr, '.');
    if (idx == string::npos || idx >= tstr.size()) // 全桁数のみ指定
    {
      vnf.first = strtol(tstr.c_str(), &ptr, 0);
      if (vnf.first == 0 || abs(vnf.first) > LENbuf) return -3; // 全桁数指定不適
    }
    else                        // 小数部つき指定
    {
      vnf.first = strtol(substr.c_str(), &ptr, 0);
      if (vnf.first == 0 || abs(vnf.first) > LENbuf) return -3;
      tstr.erase(0, idx + 1);
      vnf.second = strtol(tstr.c_str(), &ptr, 0);
      if (vnf.second == 0 && ptr == tstr.c_str()) return -4; // 整数変換できず
      if (vnf.second < 0 || vnf.second > abs(vnf.first) - 2) return -4; // 小数部桁指定不適
    }
  }
  return 0;
}

int dans::getfmtlst(list<format>& fmtlst, const string& sstr, char delmc)
// 文字列から書式を読み込みリストに追加する
{
  lstr fmtstrlst;
  if (fmtsplit(fmtstrlst, sstr, delmc) < 1) return 0;
  int size0 = fmtlst.size();
  for (lstr_itr pos = fmtstrlst.begin(); pos != fmtstrlst.end(); ++pos)
  {
    list<format> tfmtlst;
    string substr;
    strsize_t idx = string::npos;
    if ((*pos)[0] == '(')       // (書式)形式
    {
      pos->erase(0, 1);
      if ((idx = getsubstr(substr, *pos, ')')) == string::npos) return -1; // 書式不適
      if (getfmtlst(tfmtlst, substr, delmc) < 0) return -2;
      pos->erase(0, idx + 1);
      idx = getsubstr(substr, *pos, CHRrep);
    }
    else                        // 通常形式
    {
      idx = getsubstr(substr, *pos, CHRrep);
      format fmt;
      if (fmt.getfmt(substr) < 0) return -3;
      tfmtlst.push_back(fmt);
    }
    int r = 1;
    if (idx != string::npos && idx < pos->size()) // 繰り返し指定あり
    {
      pos->erase(0, idx + 1);
      if ((r = atoi(pos->c_str())) < 1) return -3;
    }
    for (int i = 0; i < r; ++i)
    {
      list<format> ttfmtlst(tfmtlst);
      fmtlst.splice(fmtlst.end(), ttfmtlst);
    }
  }
  return fmtlst.size() - size0; // 追加した書式数を返す
}
