// ftable.C Version 7.40  Copyleft (c) 杉本典夫
// Time-stamp: <2022-12-12 11:13:55 tac>
//
// 度数表クラスのメンバー関数
//
// ライブラリ:libdans.so.7.4
// コンパイル法:g++ -O -fPIC -c ftable.C

#include <dans.H>               // dans用ヘッダファイル(fun.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

namespace                       // 名前無し名前空間(局所別名の定義)
{
  using lfrow = list<frow>;     // 行リストの短名定義
  using lfrow_itr = list<frow>::iterator; // 反復子の短名定義
  using lfrow_citr = list<frow>::const_iterator; // 反復子の短名定義

  constexpr int Wpr = 5;        // %の出力桁数
  constexpr int Lmax = LENcrt - 2; // 出力行数最大値

  int put_ftabhed(ostream& fout, bool& puttotal, size_t& jend, size_t joffset, bool putpr, const ftable& ftab, tabtype tabt)
  // 度数表の列コード行を出力する
  {
    int prlen = putpr ? Wpr + 2 : 0; // %の出力幅
    int npad = ftab.hed().size() - strclen(ftab.hed(), CHARwide);
    fout << "\n" << left << setw(ftab.hwide() + npad) << ftab.hed() << right;
    int w = ftab.hwide();
    size_t j = joffset;
    while (j <= ftab.nc() && (ftab.hwide() + ftab.cwide(j) + prlen >= Lmax || w + ftab.cwide(j) + prlen < Lmax))
    {
      npad = ftab.ccode(j).size() - strclen(ftab.ccode(j), CHARwide);
      fout << setw(ftab.cwide(j) + npad) << ftab.ccode(j); // 列コード出力
      w += ftab.cwide(j);
      if (putpr)                // %出力あり
      {
        fout << "(" << setw(Wpr) << "%  ";
        fout << ")";
        w += prlen;
      }
      ++j;
    }
    if (j > ftab.nc() && (ftab.hwide() + ftab.twide() + prlen >= Lmax || w + ftab.twide() + prlen < Lmax))
    {
      string tstr("合計");
      npad = tstr.size() - strclen(tstr, CHARwide);
      fout << setw(ftab.twide() + npad) << tstr; // 行合計欄出力
      w += ftab.twide();  puttotal = true;
      if (putpr)                // %出力あり
      {
        fout << "(" << setw(Wpr) << "%  ";
        fout << ")";
        w += prlen;
      }
    }
    fout << "\n";
    jend = j - 1;               // 出力した最後の列インデックス
    return w;                   // 出力した桁数を返す
  }

  void put_ftabf(ostream& fout, size_t joffset, size_t jend, bool puttotal, bool putpr, const ftable& ftab, tabtype tabt)
  // 度数表の行コードと度数行を出力する
  {
    for (size_t i = 1; i <= ftab.nr(); ++i)
    {
      int npad = ftab.rcode(i).size() - strclen(ftab.rcode(i), CHARwide);
      fout << left << setw(ftab.hwide() + npad) << ftab.rcode(i) << right;
      for (size_t j = joffset; j <= jend; ++j)
      {
        fout << setw(ftab.cwide(j)) << ftab.f(i, j); // 度数出力
        if (putpr)                                   // %出力
        {
          double pr = 0.0;
          int ftotal = 0;
          switch (tabt)
          {
          case tabfprc : ftotal = ftab.fctotal(j); // 列合計に対する%
            break;
          case tabfprt : ftotal = ftab.ftotal(); // 全合計に対する%
            break;
          default : ftotal = ftab.frtotal(i); // 列合計に対する%
          }
          if (ftotal > 0) pr = (double)(ftab.f(i, j)) / (double)(ftotal);
          int prec0 = fout.precision(); // 現在の小数点以下の精度を保存する
          fout.setf(ios::fixed);
          fout << "(" << setw(Wpr) << setprecision(1) << pr * 100.0;
          fout << setprecision(prec0) << ")";
          fout.unsetf(ios::fixed);
        }
      }
      if (puttotal)
      {
        fout << setw(ftab.twide()) << ftab.frtotal(i);
        if (putpr)              // 行合計の%出力
        {
          double pr = 1.0;
          if (tabt != tabfprr) pr = (ftab.ftotal() > 0) ? (double)(ftab.frtotal(i)) / (double)(ftab.ftotal()) : 0.0;
          int prec0 = fout.precision(); // 現在の小数点以下の精度を保存する
          fout.setf(ios::fixed);
          fout << "(" << setw(Wpr) << setprecision(1) << pr * 100.0;
          fout << setprecision(prec0) << ")";
          fout.unsetf(ios::fixed);
        }
      }
      fout << "\n";
    }
  }

  void put_ftabtotal(ostream& fout, size_t joffset, size_t jend, bool puttotal, bool putpr, const ftable& ftab, tabtype tabt)
  // 度数表の合計行を出力する
  {
    string tstr("合計");
    int npad = tstr.size() - strclen(tstr, CHARwide);
    fout << left << setw(ftab.hwide() + npad) << tstr << right;
    for (size_t j = joffset; j <= jend; ++j)
    {
      fout << setw(ftab.cwide(j)) << ftab.fctotal(j);
      if (putpr)
      {
        double pr = 1.0;
        if (tabt != tabfprc) pr = (ftab.ftotal() > 0) ? (double)(ftab.fctotal(j)) / (double)(ftab.ftotal()) : 0.0;
        int prec0 = fout.precision(); // 現在の小数点以下の精度を保存する
        fout.setf(ios::fixed);
        fout << "(" << setw(Wpr) << setprecision(1) << pr * 100.0;
        fout << setprecision(prec0) << ")";
        fout.unsetf(ios::fixed);
      }
    }
    if (puttotal)
    {
      fout << setw(ftab.twide()) << ftab.ftotal();
      if (putpr)                // 行合計の%出力
      {
        double pr = (ftab.ftotal() > 0) ? 1.0 : 0.0;
        int prec0 = fout.precision(); // 現在の小数点以下の精度を保存する
        fout.setf(ios::fixed);
        fout << "(" << setw(Wpr) << setprecision(1) << pr * 100.0;
        fout << setprecision(prec0) << ")";
        fout.unsetf(ios::fixed);
      }
    }
  }
} // 名前無し名前空間終わり

const frow& frow::operator=(const frow& frow0) // コピー代入演算子
{
  if (&frow0 == this) return *this; // 自分自身への代入を防ぐ
  vrcode = frow0.vrcode;  vfvec = frow0.vfvec;  vcum = frow0.vcum;
  return *this;
}

const frow& frow::operator=(int f0) // 度数代入演算子(全データ同値代入)
{
  vfvec.assign(vfvec.size(), f0);
  vfvec[0] = f0 * nc();  vcum = false;
  return *this;
}

const frow& frow::operator+=(const frow& frow0) // 加算演算子
{
  if (nc() != frow0.nc()) return *this; // 行数が同じ時だけ加算する
  for (size_t j = 0; j < vfvec.size(); ++j) vfvec[j] += frow0.vfvec[j];
  return *this;
}

void frow::clear()              // 行をクリアする(初期状態にする)
{
  vrcode.erase();  vfvec.resize(1);
  vfvec[0] = 0;  vcum = false;
}

size_t frow::erase(size_t cno)  // 行要素を削除する
{
  if (cno < 1 || cno > nc()) return nc(); // 番号指定不適
  vfvec[0] -= vfvec[cno];
  for (size_t j = cno; j < nc(); ++j) vfvec[j] = vfvec[j + 1];
  vfvec.resize(vfvec.size() - 1);
  return nc();                  // 削除後の行数を返す
}

size_t frow::erase()   // 全ての行要素を削除する(行コードは削除しない)
{
  vfvec.resize(1);
  vfvec[0] = 0;  vcum = false;
  return nc();                  // 削除後の行数を返す
}

size_t frow::resize(size_t nc0) // 列数を変更する
{
  size_t nctmp = vfvec.size() - 1;
  if (nc0 == nctmp) return nc0; // 変更なし
  if (nc0 > nctmp)              // 現在の列数よりも大きい時
  { vfvec.resize(nc0 + 1, 0); }
  else                          // 現在の列数よりも小さい時
  {
    for (size_t j = nc0 + 1; j <= nctmp; ++j) vfvec[0] -= vfvec[j];
    vfvec.resize(nc0 + 1);
  }
  return nc0;                   // 変更後の列数を返す
}

int frow::assign(size_t cno, int f0) // 度数を設定する
{
  if (cno < 1 || cno > nc()) return -1; // 変更できず
  if (f0 == vfvec[cno]) return f0; // 変更せず
  vfvec[0] += f0 - vfvec[cno];
  vfvec[cno] = f0;
  return f0;
}

bool frow::fcum()               // 累積度数表に変換する
{
  if (vcum || vfvec.size() < 3) return vcum; // 必要なし
  int cum = 0;
  for (size_t j = 1; j < vfvec.size(); ++j) // 行合計はそのまま
  {
    cum += vfvec[j];
    vfvec[j] = cum;
  }
  vcum = true;
  return vcum;
}

int frow::get(datac& cdata)     // dansデータから度数を読み込む
{
  if (vfvec.size() < 2) vfvec.resize(cdata.size());
  vrcode = cdata.sdata(0);
  vfvec[0] = 0;  vcum = false;
  size_t n = min(vfvec.size(), cdata.size());
  for (size_t j = 1; j < n; ++j)
  {
    double fdata = cdata.fdata(j);
    vfvec[j] = (fdata == BLANKf) ? 0 : (int)fdata; // 欠損は0にする
    vfvec[0] += vfvec[j];
  }
  for (size_t j = n; j < vfvec.size(); ++j) vfvec[j] = 0; // 残りは0にする
  return n;                     // 読み込んだデータ数を返す
}

ftable::ftable() : frowlst(), dcode(), vccodevec(1, ""), vcwidevec(2, 0), ritrvec(1)
// デフォールトコンストラクタ
{
  frow ctotal(0, "合計");
  frowlst.push_back(ctotal);
  ritrvec[0] = frowlst.begin();
}

ftable::ftable(size_t nr0, size_t nc0) : frowlst(), dcode(), vccodevec(nc0 + 1, ""), vcwidevec(nc0 + 2, 0), ritrvec(nr0 + 1)
// コンストラクタ
{
  frow ctotal(nc0, "合計"), row(nc0, "");
  frowlst.push_back(ctotal);
  for (size_t i = 1; i <= nr0; ++i) frowlst.push_back(row);
  size_t i = 0;
  for (lfrow_itr fpos = frowlst.begin(); fpos != frowlst.end(); ++fpos, ++i) ritrvec[i] = fpos;
  // 反復子ベクトルをセットする
}

ftable::ftable(const ftable& ftable0) : frowlst(ftable0.frowlst), dcode(ftable0.dcode), vccodevec(ftable0.vccodevec), vcwidevec(ftable0.vcwidevec), ritrvec(ftable0.ritrvec)
// コピーコンストラクタ
{
  size_t i = 0;
  for (lfrow_itr fpos = frowlst.begin(); fpos != frowlst.end(); ++fpos, ++i) ritrvec[i] = fpos;
  // 反復子ベクトルをセットする
}

const ftable& ftable::operator=(const ftable& ftable0) // コピー代入演算子
{
  if (&ftable0 == this) return *this; // 自分自身への代入を防ぐ
  frowlst = ftable0.frowlst;  dcode = ftable0.dcode;
  vccodevec = ftable0.vccodevec;
  vcwidevec = ftable0.vcwidevec;  ritrvec = ftable0.ritrvec;
  size_t i = 0;
  for (lfrow_itr fpos = frowlst.begin(); fpos != frowlst.end(); ++fpos, ++i) ritrvec[i] = fpos;
  // 反復子ベクトルをセットする
  return *this;
}

const ftable& ftable::operator=(int f0) // 度数代入演算子(全度数同値代入)
{
  lfrow_itr fpos = frowlst.begin();
  lfrow_itr fpostotal = fpos;
  *fpostotal = 0;               // 度数合計行を0にする
  ++fpos;
  while (fpos != frowlst.end())
  {
    *fpos = f0;                 // 行の全度数にf0を代入
    *fpostotal += *fpos;        // 合計行に度数を加算する
    ++fpos;
  }
  return *this;
}

void ftable::clear()            // 度数表をクリアする(初期状態にする)
{
  frowlst.clear();  ritrvec.resize(1);
  frow ctotal(0, "合計");
  frowlst.push_back(ctotal);
  ritrvec[0] = frowlst.begin();
  dcode.erase();  vccodevec.resize(1);  vcwidevec.resize(2);
  vccodevec[0] = "";  vcwidevec[0] = vcwidevec[1] = 0;
}

size_t ftable::rerase(size_t rno) // 行を削除する
{
  if (rno < 1 || rno > nr()) return nr(); // 番号指定不適
  frowlst.erase(ritrvec[rno]);
  set();
  return nr();                  // 削除後の行数を返す
}

size_t ftable::rerase()    // 全ての行を削除する(列コードは削除しない)
{
  frowlst.clear();  ritrvec.resize(1);
  frow ctotal(nc(), "合計");
  frowlst.push_back(ctotal);
  ritrvec[0] = frowlst.begin();
  return nr();                  // 削除後の行数を返す
}

size_t ftable::cerase(size_t cno) // 列を削除する
{
  if (cno < 1 || cno > nc()) return nc(); // 番号指定不適
  for (size_t j = cno; j < nc(); ++j)
  {
    vccodevec[j] = vccodevec[j + 1];
    vcwidevec[j] = vcwidevec[j + 1];
  }
  vcwidevec[nc()] = vcwidevec[nc() + 1];
  vccodevec.resize(vccodevec.size() - 1);
  vcwidevec.resize(vcwidevec.size() - 1);
  for (lfrow_itr fpos = frowlst.begin(); fpos != frowlst.end(); ++fpos) fpos->erase(cno);
//  for (auto frowi : frowlst) frowi.erase(cno);
  return nc();                  // 削除後の列数を返す
}

size_t ftable::cerase()         // 全ての列を削除する
{
  frowlst.clear();  ritrvec.resize(1);
  frow ctotal(0, "合計");
  frowlst.push_back(ctotal);
  ritrvec[0] = frowlst.begin();
  vccodevec.resize(1);  vcwidevec.resize(2);
  vcwidevec[1] = 0;
  return nc();                  // 削除後の列数を返す
}

int ftable::resize(size_t nr0, size_t nc0) // 行数・列数を変更する
{
  if (nr0 == nr() && nc0 == nc()) return 0; // 変更せず
  if (nc0 != nc())              // 列数を変更する
  {
    size_t nctmp = nc();
    vccodevec.resize(nc0 + 1, "");
    int twide = vcwidevec[nctmp + 1];
    vcwidevec.resize(nc0 + 2, 0);
    vcwidevec[nc0 + 1] = twide;
    for (lfrow_itr fpos = frowlst.begin(); fpos != frowlst.end(); ++fpos) fpos->resize(nc0);
//    for (auto frowi : frowlst) frowi.resize(nc0);
  }
  if (nr0 != nr())              // 行数を変更する
  {
    frow row(nc0, "");
    frowlst.resize(nr0 + 1, row);
  }
  set();
  return 1;                     // 変更
}

int ftable::assign(size_t rno, size_t cno, int f0) // 度数を設定する
{
  if (rno < 1 || rno > nr() || cno < 1 || cno > nc()) return -1;
  if (f0 == ritrvec[rno]->f(cno)) return f0; // 変更せず
  int ctotal = ritrvec[0]->f(cno) - ritrvec[rno]->f(cno) + f0;
  ritrvec[rno]->assign(cno, f0);  ritrvec[0]->assign(cno, ctotal);
  return f0;
}

int ftable::get(datac& cdata)   // 度数表を読み込む(整理はしない)
{
  constexpr int Wdef = 6;       // 省略時桁数
  if (nc() < 1)                 // 度数表0行目(列コード読み込み)
  {
    vccodevec.resize(cdata.size());
    vcwidevec.resize(cdata.size() + 1);
    for (size_t j = 0; j < cdata.size(); ++j)
    {
      vccodevec[j] = cdata.sdata(j);
      vcwidevec[j] = cdata.sdata(j).size();
    }
    vcwidevec[cdata.size()] = Wdef; // 行合計欄の省略時桁数
    (frowlst.begin())->resize(cdata.size() - 1); // 度数合計行の行数を変更する
  }
  else                          // 度数表1行目以降(度数読み込み)
  {
    frow row(nc(), cdata.sdata(0));
    row.get(cdata);
    frowlst.push_back(row);
  }
  return cdata.size();          // データ数を返す
}

size_t ftable::set()            // 度数表を整理し行数を返す
{
  if (nc() < 1) return 0;       // 度数表なし
  ritrvec.resize(frowlst.size());
  int w = strclen(vccodevec[0], CHARwide);
  int wmax = max(4, w);
  lfrow_itr fpos = frowlst.begin();
  lfrow_itr fpostotal = fpos;
  ritrvec[0] = fpos;
  *fpostotal = 0;               // 合計行の度数を0にする
  ++fpos;
  size_t i = 1;
  while (fpos != frowlst.end())
  {
    w = strclen(fpos->rcode(), CHARwide);
    if (w > wmax) wmax = w;
    ritrvec[i] = fpos;          // 反復子ベクトルをセットする
    *fpostotal += *fpos;        // 合計行に度数を加算する
    ++fpos;  ++i;
  }
  vcwidevec[0] = wmax;          // ヘッダの桁数
  for (size_t j = 1; j <= nc(); ++j)
  {
    int wf = (int)(log10((double)(max(1, ritrvec[0]->f(j))))) + 2; // 度数の出力桁数
    w = strclen(vccodevec[j], CHARwide);
    vcwidevec[j] = max(wf, w);
  }
  w = (int)(log10((double)(max(1, ritrvec[0]->ftotal())))) + 2;
  vcwidevec[nc() + 1] = max(6, w); // 行合計の桁数
  return nr();                  // 行数を返す
}

int ftable::trm()               // 合計が0の行・列を削除する
{
  if (nc() < 1 || nr() < 1) return 0; // 度数表なし
  int erase = 0;
  lfrow_itr fpos = frowlst.begin();
  ++fpos;                       // 合計行をとばす
  while (fpos != frowlst.end())
  {
    if (fpos->ftotal() < 1) { fpos = frowlst.erase(fpos);  erase = 1; }
    else ++fpos;
  }
  if (erase) set();             // 行を削除した時は度数表を整理する
  size_t n = 0;
  for (size_t j = 1; j <= nc(); ++j)
  {
    if (ritrvec[0]->f(j) > 0)
    {
      ++n;
      if (n < j)                // 削除した時だけ移動する
      {
        vccodevec[n] = vccodevec[j];  vcwidevec[n] = vcwidevec[j];
        for (size_t i = 0; i <= nr(); ++i) ritrvec[i]->assign(n, ritrvec[i]->f(j));
      }
    }
  }
  if (n < nc()) { nc(n);  erase += 2; }
  return erase;    // 0:削除なし,1:行のみ削除,2:列のみ削除,3:両方削除
}

bool ftable::fcum()             // 累積度数表に変換する
{
  if (ritrvec[0]->cum() || nc() < 2) return ritrvec[0]->cum(); // 必要なし
  for (size_t i = 0; i <= nr(); ++i) ritrvec[i]->fcum(); // 列合計も変換する
  return ritrvec[0]->cum();
}

int dans::fgetftable(string& fname, listidd& idlst, listftab& ftablst)
// データ識別情報と度数集計表を読み込みリストに保存して返す
{
  istream* finp = nullptr;
  if (fname.empty())
  { finp = &cin;  fname = "stdin"; }
  else
  {
    finp = new ifstream(fname.c_str());
    if (!(*finp)) return disperr(fname, "ファイルがありません");
  }
  string sline;                 // データ入力用
  datac cdata;                  // 度数表入力用
  idd id;                       // 個々のデータ識別情報
  ftable ftab;                  // 個々の度数表
  idlst.clear();  ftablst.clear();
  while (getsline(sline, *finp) != EOF)
  {
    strchomp(sline);
    if (sline[0] == CHRtags)    // データ識別情報
    {
      if (ftab.nc() > 0)
      {
        if (id.ni() < 1) id.set(fname, 1);
        ftab.set();
        idlst.push_back(id);  ftablst.push_back(ftab);
        cdata.clear();  id.clear();  ftab.clear();
      }
      if (id.append(sline) < 0) return disperr(fname, string("データ識別情報が間違っています→") + sline);
    }
    else if (sline.empty())     // 改行のみの行
    {
      if (id.ni() > 0 || ftab.nc() > 0)
      {
        if (id.ni() < 1) id.set(fname, 1);
        ftab.set();
        idlst.push_back(id);  ftablst.push_back(ftab);
        cdata.clear();  id.clear();  ftab.clear();
      }
    }
    else                        // 度数データ読み込み
    {
      cdata.getdata(sline);
      ftab.get(cdata);
    }
  }
  if (id.ni() > 0 || ftab.nc() > 0)
  {
    if (id.ni() < 1) id.set(fname, 1);
    ftab.set();
    idlst.push_back(id);  ftablst.push_back(ftab);
  }
  if (finp != &cin) delete finp;
  return (int)(ftablst.size()); // 読み込んだ度数表数を返す
}

int dans::putftable(ostream& fout, const ftable& ftab, tabtype tabt)
// 度数表を度数表形式で出力する
{
  bool putpr = (tabt == tabfprc || tabt == tabfprr || tabt == tabfprt) ? true : false;
  bool puttotal = false;        // 行合計出力フラッグ
  size_t joffset = 1, jend = 0;
  int nline = 0;                // 出力行数
  while (!puttotal)             // 表をLmax桁以内に分割して出力する
  {
    int w = put_ftabhed(fout, puttotal, jend, joffset, putpr, ftab, tabt);
    putline(fout, w);
    put_ftabf(fout, joffset, jend, puttotal, putpr, ftab, tabt);
    putline(fout, w);
    put_ftabtotal(fout, joffset, jend, puttotal, putpr, ftab, tabt);
    fout << endl;
    joffset = jend + 1;  nline += ftab.nr() + 4;
  }
  return nline;                 // 出力行数を返す
}
